r"""PyTorch Image Quality Assessement

The piqa package is divided in several submodules, each of
which implements the functions and/or classes related to a
specific image quality assessement metric.
"""

__version__ = '1.0.3'
